#include "matrix.h"

// création d’une nouvelle matrice de m lignes et n colonnes et allocation de la mémoire
error_code matrix_alloc(matrix *mat, int32_t m, int32_t n){
     if (mat == NULL) return uninitialized;
     if (m <= 0 || n <= 0) return uninitialized;

     mat->m = m;
     mat->n = n;

     //alloue un tableau de taille mat->m
     mat->data = malloc(mat->m * sizeof(int32_t *));

     //si le malloc echoue
     if (mat->data == NULL) {
          return memory_error;
     }

     for (int32_t i = 0; i < mat->m; i++) {
          //alloue un table de taille mat->n
          mat->data[i] = malloc(mat->n * sizeof(int32_t));
          //si le malloc echoue pour chaque valeur
          if (mat->data[i] == NULL) {
               for (int32_t j = 0; j < i; j++) {
                    free(mat->data[j]);
               }
               free(mat->data);
               return memory_error;
          }
     }

     return ok;
}

// allocation et initialisation à une valeur, val, d’une nouvelle matrice de m lignes et n colonnes
error_code matrix_init(matrix *mat, int32_t m, int32_t n, int32_t val){
     matrix_alloc(mat, m, n);

     if (m <= 0 || n <= 0 || val <= 0) return uninitialized;
     if (m > mat->m || n > mat->n) return out_of_bounds;

     for (int32_t i = 0; i < mat->m; i++) {
          for (int32_t j = 0; j < mat->n; j++) {
               mat->data[i][j] = val;
          }
     }

     return ok;
}

// libération de la mémoire de la matrice en argument, le pointeur vers les données est mis à NULL, le nombre de lignes et de colonnes sont mis à -1
error_code matrix_destroy(matrix *mat){
     if (mat == NULL || mat->data == NULL) return uninitialized;

     for (int32_t i = 0; i < mat->m; i++) {
          free(mat->data[i]);
     }
     free(mat->data);
     mat->data = NULL;

     return ok;
}

// allocation d’une matrice, et initialisation de ses valeurs à partir d’un tableau de taille s = m*n
error_code matrix_init_from_array(matrix *mat, int32_t m, int32_t n, int32_t data[], int32_t s){
     int count = 0;
     if(m == 0 || n == 0){
          return uninitialized;
     }

     if (m * n != s) {
          return out_of_bounds;
     }

     matrix_alloc(mat, m, n);

     for (int32_t i = 0; i < mat->m; i++) {
          for (int32_t j = 0; j < mat->n; j++) {
               mat->data[i][j] = data[count];
               count++;
               if (count > s) {
                    return ok;
               }
          }
     }

     return ok;
}

// création du clone d’une matrice, la nouvelle matrice est une copie de la matrice d’origine
error_code matrix_clone(matrix *cloned, const matrix mat){
     matrix_alloc(cloned, mat.m, mat.n);

     if (cloned == NULL || mat.data == NULL) return uninitialized;

     for (int i = 0; i < mat.m; i++) {
          for (int j = 0; j < mat.n; j++) {
               cloned->data[i][j] = mat.data[i][j];
          }
     }


     return ok;
}

// création de la matrice transposée d’une matrice, la nouvelle matrice est une copie de la matrice originale ou les lignes et les colonnes sont échangées
error_code matrix_transpose(matrix *transposed, const matrix mat){
     matrix_alloc(transposed, mat.n, mat.m);

     if (transposed == NULL || transposed->data == NULL || mat.data == NULL) return uninitialized;
     if (transposed->m != mat.n || transposed->n != mat.m) return out_of_bounds;

     for (int32_t i = 0; i < transposed->m; i++) {
          for (int32_t j = 0; j < transposed->n; j++) {
               transposed->data[i][j] = mat.data[j][i];
          }
     }

     return ok;
}

// affichage d’une matrice (très utile pour le débogage)
error_code matrix_print(const matrix mat){
     int count = 0;
     for (int32_t i = 0; i < mat.m; i++) {
          for (int32_t j = 0; j < mat.n; j++) {
               printf("%d ", mat.data[i][j]);
               count++;

               if (count == mat.n) {
                    printf("\n");
                    count = 0;
               }
          }
     }
     return ok;
}

// extraction d’une sous-matrice d’une matrice
// 1 3 0 2 -> m = 2 n = 2
error_code matrix_extract_submatrix(matrix *sub, const matrix mat, int32_t m0, int32_t m1, int32_t n0, int32_t n1){
     matrix_alloc(sub, m1 - m0 + 1, n1 - n0 + 1);

     if (sub == NULL || sub->data == NULL || mat.data == NULL) {
          return uninitialized;
     }
     if (m0 > m1 || n0 > n1) {
          return out_of_bounds;
     }
     if (m1 > mat.m || n1 > mat.n) {
          return out_of_bounds;
     }

     if ((m1 - m0 > sub->m) || (n1 - n0 > sub->n)) {
          return out_of_bounds;
     }

     for (int i = m0; i < m1; i++) {
          for (int j = n0; j < n1; j++) {
               sub->data[i - m0][j - n0] = mat.data[i][j];
          }
     }
     return ok;
}

// test de l’égalité de deux matrices
bool matrix_is_equal(matrix mat1, matrix mat2){
     if (mat1.m == mat2.m && mat1.n == mat2.n) {
          for (int32_t i = 0; i < mat1.m; i++) {
               for (int32_t j = 0; j < mat1.n; j++) {
                    if (mat1.data[i][j] != mat2.data[i][j]) {
                         return false;
                    }else{
                         return true;
                    }
               }
          }
     }else{
          return false;
     }
     return true;
}

// récupération de l’élément [ix][iy] de la matrice de façon sûre (vérification des dépassements de capacité par exemple) et copie dans elem
error_code matrix_get(int32_t *elem, const matrix mat, int32_t ix, int32_t iy){
     if (ix <= mat.m && iy <= mat.n) {
          *elem = mat.data[ix][iy];
          return ok;
     }else{
          return out_of_bounds;
     }

     return ok;
}

// modification d’un élément [ix][iy] de la matrice de façon sûre (vérification des dépassements de capacité par exemple)
error_code matrix_set(matrix *mat, int32_t ix, int32_t iy, int32_t elem){
     if (ix <= mat->m && iy <= mat->n) {
          mat->data[ix][iy] = elem;
          return ok;
     }else{
          return out_of_bounds;
     }

     return ok;
}

// appliquer une fonction sur chaque élément d’une matrice “en place” (in place en bon français)
error_code matrix_map_ip(matrix *mat, void (*foo)(int *)){
     for (int32_t i = 0; i < mat->m; i++) {
          for (int32_t j = 0; j < mat->n; j++) {
               foo(&mat->data[i][j]);
          }
     }
     return ok;
}

// appliquer une fonction sur chaque élément d’une matrice et enregistrer le résultat dans une nouvelle matrice
error_code matrix_map(matrix *mapped, const matrix mat, void (*foo)(int *)){
     if (mapped == NULL) return uninitialized;

     matrix_clone(mapped, mat);

     for (int32_t i = 0; i < mapped->m; i++) {
          for (int32_t j = 0; j < mapped->n; j++) {
               foo(&mapped->data[i][j]);
          }
     }

     return ok;
}
